#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
import pytest

from pyomo.environ import units

from idaes.core.base.property_set import (
    PropertyMetadata,
    PropertySetBase,
    StandardPropertySet,
    ElectrolytePropertySet,
    _PropertyMetadataIndex,
)
from idaes.core.base.property_meta import UnitSet
from idaes.core.util.exceptions import PropertyPackageError


class DummyMeta:
    pass


class Test_PropertyMetadataIndex:
    @pytest.fixture(scope="function")
    def meta(self):
        p = DummyMeta()

        meta = _PropertyMetadataIndex(parent=p, idx="baz")

        return meta

    @pytest.mark.unit
    def test_all_args(self, meta):
        p = DummyMeta()

        meta = _PropertyMetadataIndex(
            parent=p,
            idx="bar",
            method="foo",
            supported=True,
            required=True,
            valid_range=(0, 1),
        )
        assert meta.method == "foo"
        assert meta._parent is p
        assert meta._idx == "bar"
        assert meta.supported
        assert meta.required
        assert meta.valid_range == (0, 1)

    @pytest.mark.unit
    def test_default_value(self, meta):
        assert meta.method is None
        assert not meta.supported
        assert not meta.required

        assert meta._idx == "baz"
        assert meta.valid_range is None

    @pytest.mark.unit
    def test_set_method(self, meta):
        assert meta.method is None
        meta.set_method("foo")
        assert meta.method == "foo"

    @pytest.mark.unit
    def test_set_required(self, meta):
        assert not meta.required
        meta.set_required(True)
        assert meta.required

    @pytest.mark.unit
    def test_set_supported(self, meta):
        assert not meta.supported
        meta.set_supported(True)
        assert meta.supported

    @pytest.mark.unit
    def test_valid_range_verification_invalid(self):
        p = DummyMeta()

        with pytest.raises(
            ValueError, match=r"valid_range must be a tuple of length 2 \(got foo\)"
        ):
            meta = _PropertyMetadataIndex(
                parent=p,
                idx="bar",
                method="foo",
                supported=True,
                required=True,
                valid_range="foo",
            )

    @pytest.mark.unit
    def test_valid_range_verification_length(self):
        p = DummyMeta()

        with pytest.raises(
            ValueError,
            match=r"valid_range must be a tuple of length 2 \(got \(1, 2, 3\)\)",
        ):
            meta = _PropertyMetadataIndex(
                parent=p,
                idx="bar",
                method="foo",
                supported=True,
                required=True,
                valid_range=(1, 2, 3),
            )

    @pytest.mark.unit
    def test_valid_range_verification_values(self):
        p = DummyMeta()

        with pytest.raises(
            ValueError,
            match=r"valid_range must be a 2-tuple with form \(lower, upper\): first value "
            r"was greater than second value: \(1, 0\)",
        ):
            meta = _PropertyMetadataIndex(
                parent=p,
                idx="bar",
                method="foo",
                supported=True,
                required=True,
                valid_range=(1, 0),
            )

    @pytest.mark.unit
    def test_update_property_args(self, meta):
        assert meta.method is None
        assert not meta.supported
        assert not meta.required
        assert meta.valid_range is None

        meta.update_property(method="foo", required=True, supported=False)

        assert meta.method == "foo"
        assert not meta.supported
        assert meta.required
        assert meta.valid_range is None

        meta.update_property(
            method="foo", required=False, supported=True, valid_range=(0, 10)
        )

        assert meta.method == "foo"
        assert meta.supported
        assert not meta.required
        assert meta.valid_range == (0, 10)


class TestPropertyMetadata:
    @pytest.fixture(scope="function")
    def meta(self):
        meta = PropertyMetadata(
            name="test",
            units=units.dimensionless,
        )

        return meta

    @pytest.mark.unit
    def test_no_name(self):
        with pytest.raises(TypeError, match='"name" is required'):
            meta = PropertyMetadata(
                name=None,
                units=units.dimensionless,
            )

    @pytest.mark.unit
    def test_all_args(self):
        meta = PropertyMetadata(
            name="test",
            units=units.dimensionless,
            indices=["a", "b", "none"],
            initialize={
                None: {"method": "foo", "required": True, "supported": True},
                "a": {"method": "bar", "required": False, "supported": True},
                "b": {"method": "baz", "required": True, "supported": False},
            },
        )

        assert meta.name == "test"
        assert meta.units is units.dimensionless
        assert meta._indices == ["a", "b", "none"]

        assert isinstance(meta._none, _PropertyMetadataIndex)
        assert meta._none.method == "foo"
        assert meta._none.required
        assert meta._none.supported

        assert isinstance(meta._a, _PropertyMetadataIndex)
        assert meta._a.method == "bar"
        assert not meta._a.required
        assert meta._a.supported

        assert isinstance(meta._b, _PropertyMetadataIndex)
        assert meta._b.method == "baz"
        assert meta._b.required
        assert not meta._b.supported

    @pytest.mark.unit
    def test_default_value(self, meta):
        assert meta.name == "test"
        assert meta.units is units.dimensionless
        assert meta._indices is None
        assert isinstance(meta._none, _PropertyMetadataIndex)


class TestPropertySetBase:
    @pytest.fixture(scope="function")
    def pset(self):
        p = DummyMeta()
        p.default_units = "foo"

        pset = PropertySetBase(parent=p)

        return pset

    @pytest.mark.unit
    def test_expected_attrs(self):
        p = object()
        pset = PropertySetBase(parent=p)

        assert pset._parent_block is p
        assert pset._defined_properties == []
        assert pset._defined_indices == ["comp", "phase", "phase_comp"]

    @pytest.mark.unit
    def test_direct_assignment(self):
        p = object()
        pset = PropertySetBase(parent=p)

        with pytest.raises(
            TypeError,
            match="PropertySets do not support direct assignment. Please use define_property",
        ):
            pset.foo = "bar"

    @pytest.mark.unit
    def test_add_property(self, pset):
        pset._add_property(
            name="foo",
            method="baz",
            supported=True,
            required=True,
            units=units.dimensionless,
        )

        assert isinstance(pset.foo, PropertyMetadata)
        assert pset.foo.name == "foo"
        assert pset.foo._none.method == "baz"
        assert pset.foo._none.supported
        assert pset.foo._none.required
        assert pset.foo.units is units.dimensionless
        for i in pset.foo._indices:
            assert i in ["comp", "phase", "phase_comp", "none"]

        assert "foo" in pset._defined_properties

        assert isinstance(pset.foo._none, _PropertyMetadataIndex)
        assert isinstance(pset.foo._comp, _PropertyMetadataIndex)
        assert isinstance(pset.foo._phase, _PropertyMetadataIndex)
        assert isinstance(pset.foo._phase_comp, _PropertyMetadataIndex)

    @pytest.mark.unit
    def test_add_property_unindexed_init(self, pset):
        pset._add_property(
            name="foo",
            initialize={
                "none": {
                    "method": "baz",
                    "supported": True,
                    "required": True,
                },
            },
            units=units.dimensionless,
        )

        assert isinstance(pset.foo, PropertyMetadata)
        assert pset.foo.name == "foo"
        assert pset.foo._none.method == "baz"
        assert pset.foo._none.supported
        assert pset.foo._none.required
        assert pset.foo.units is units.dimensionless
        for i in pset.foo._indices:
            assert i in ["comp", "phase", "phase_comp", "none"]

        assert "foo" in pset._defined_properties

        assert isinstance(pset.foo._none, _PropertyMetadataIndex)
        assert isinstance(pset.foo._comp, _PropertyMetadataIndex)
        assert isinstance(pset.foo._phase, _PropertyMetadataIndex)
        assert isinstance(pset.foo._phase_comp, _PropertyMetadataIndex)

    @pytest.mark.unit
    def test_add_property_w_index(self, pset):
        pset._add_property(
            name="foo",
            units=units.dimensionless,
            indices=["a", "b"],
        )

        assert isinstance(pset.foo, PropertyMetadata)
        assert pset.foo.name == "foo"
        assert pset.foo.units is units.dimensionless

        for i in pset.foo._indices:
            assert i in ["a", "b"]

        assert "foo" in pset._defined_properties

        assert isinstance(pset.foo._a, _PropertyMetadataIndex)
        assert isinstance(pset.foo._b, _PropertyMetadataIndex)
        assert not hasattr(pset.foo, "_none")

    @pytest.mark.unit
    def test_add_property_w_index_other(self, pset):
        with pytest.raises(
            ValueError,
            match="Cannot assign method, required or supported attributes directly to indexed "
            "properties. Please use the initialize argument instead.",
        ):
            pset._add_property(
                name="foo",
                method="baz",
                supported=True,
                required=True,
                units=units.dimensionless,
                indices=["a", "b"],
            )

    @pytest.mark.unit
    def test_add_property_init_and_other(self, pset):
        with pytest.raises(
            ValueError,
            match="Cannot provide values for initialize and any of method, required or supported "
            "arguments. Please use the one approach or the other.",
        ):
            pset._add_property(
                name="foo",
                method="baz",
                supported=True,
                required=True,
                units=units.dimensionless,
                initialize="bar",
            )

    @pytest.mark.unit
    def test_define_property(self, pset):
        pset.define_property(
            name="foo",
            method="baz",
            supported=True,
            required=True,
            units=units.dimensionless,
        )

        assert isinstance(pset.foo, PropertyMetadata)
        assert pset.foo.name == "foo"
        assert pset.foo._none.method == "baz"
        assert pset.foo._none.supported
        assert pset.foo._none.required
        assert pset.foo.units is units.dimensionless

        assert isinstance(pset.foo._none, _PropertyMetadataIndex)
        assert isinstance(pset.foo._comp, _PropertyMetadataIndex)
        assert isinstance(pset.foo._phase, _PropertyMetadataIndex)
        assert isinstance(pset.foo._phase_comp, _PropertyMetadataIndex)

    @pytest.mark.unit
    def test_define_property_duplicate(self, pset):
        pset.define_property(
            name="foo",
            method="baz",
            supported=True,
            required=True,
            units=units.dimensionless,
        )

        with pytest.raises(
            PropertyPackageError,
            match="A property with the name foo already exists. "
            "Please use update_property method if you wish to update "
            "an existing property's metadata.",
        ):
            pset.define_property(
                name="foo",
                method="baz",
                supported=True,
                required=True,
                units=units.dimensionless,
            )

    @pytest.mark.unit
    def test_getitem(self, pset):
        pset.define_property(
            name="FOO",
            method="baz",
            supported=True,
            required=True,
            units=units.dimensionless,
        )

        assert pset["FOO"] is pset.FOO._none
        assert pset["FOO_comp"] is pset.FOO._comp
        assert pset["FOO_phase"] is pset.FOO._phase
        assert pset["FOO_phase_comp"] is pset.FOO._phase_comp

    @pytest.mark.unit
    def test_getitem_no_present(self, pset):
        with pytest.raises(
            ValueError,
            match="Unhandled property: foo. This is mostly likely due "
            "to the property not being defined in this PropertySet.",
        ):
            pset["foo"]

    @pytest.mark.unit
    def test_iter(self, pset):
        pset.define_property(
            name="FOO",
            method="baz",
            supported=True,
            required=True,
            units=units.dimensionless,
        )
        pset.define_property(
            name="_BAR",
            method="baz",
            supported=True,
            required=True,
            units=units.dimensionless,
        )

        for i in pset:
            assert i in [pset.FOO, pset._BAR]

    @pytest.mark.unit
    def test_list_required_properties(self, pset):
        pset.define_property(
            name="FOO",
            units=units.dimensionless,
        )
        pset.FOO["none"].set_required(True)
        pset.FOO["phase"].set_required(True)

        rp = pset.list_required_properties()
        for p in rp:
            assert p in [pset.FOO[None], pset.FOO["phase"]]

    @pytest.mark.unit
    def test_check_required_properties(self, pset):
        # ADd some required properties
        pset.define_property(
            name="FOO",
            units=units.dimensionless,
        )
        pset.FOO["none"].set_required(True)
        pset.FOO["phase"].set_required(True)
        pset.FOO["phase_comp"].set_required(True)

        # Create a second property set, which only supports some required properties
        pset2 = PropertySetBase(parent=pset._parent_block)
        # In metadata but explicitly not supported
        pset2.define_property(
            name="FOO",
            units=units.dimensionless,
        )
        pset2.FOO["none"].set_supported(True)
        pset2.FOO["phase"].set_supported(True)
        pset2.FOO["phase_comp"].set_supported(False)

        unsupported = pset.check_required_properties(pset2)
        assert "FOO_phase_comp" in unsupported
        assert len(unsupported) == 1

    @pytest.mark.unit
    def test_list_supported_properties(self, pset):
        pset.define_property(
            name="FOO",
            units=units.dimensionless,
        )
        pset.FOO["none"].set_supported(True)
        pset.FOO["phase"].set_supported(True)

        sp = pset.list_supported_properties()
        for p in sp:
            assert p in [pset.FOO[None], pset.FOO["phase"]]

    @pytest.mark.unit
    def test_list_unsupported_properties(self, pset):
        pset.define_property(
            name="FOO",
            units=units.dimensionless,
        )
        pset.FOO["none"].set_supported(True)
        pset.FOO["phase"].set_supported(True)

        usp = pset.list_unsupported_properties()
        for p in usp:
            assert p in [pset.FOO["comp"], pset.FOO["phase_comp"]]

    @pytest.mark.unit
    def test_unitset(self):
        p = DummyMeta()
        p.default_units = "foo"

        pset = PropertySetBase(parent=p)

        assert pset.unitset is p.default_units

    @pytest.mark.unit
    def test_get_name_and_index(self):
        p = DummyMeta()
        p.default_units = "foo"

        pset = PropertySetBase(parent=p)
        pset._defined_properties.append("foo_bar")

        n, i = pset.get_name_and_index("foo_bar")
        assert n == "foo_bar"
        assert i is None

        n, i = pset.get_name_and_index("foo_bar_phase")
        assert n == "foo_bar"
        assert i == "phase"

        n, i = pset.get_name_and_index("foo_bar_comp")
        assert n == "foo_bar"
        assert i == "comp"

        n, i = pset.get_name_and_index("foo_bar_phase_comp")
        assert n == "foo_bar"
        assert i == "phase_comp"

        with pytest.raises(
            ValueError,
            match="Unhandled property: baz. This is mostly likely due to the "
            "property not being defined in this PropertySet.",
        ):
            pset.get_name_and_index("baz")

    @pytest.mark.unit
    def test_get_name_and_index_phase_frac(self):
        p = DummyMeta()
        p.default_units = "foo"

        pset = PropertySetBase(parent=p)
        pset._defined_properties.append("phase_frac")

        n, i = pset.get_name_and_index("phase_frac")
        assert n == "phase_frac"
        assert i is None

        n, i = pset.get_name_and_index("phase_frac_comp")
        assert n == "phase_frac"
        assert i == "comp"

        n, i = pset.get_name_and_index("phase_frac_phase")
        assert n == "phase_frac"
        assert i == "phase"

        n, i = pset.get_name_and_index("phase_frac_phase_comp")
        assert n == "phase_frac"
        assert i == "phase_comp"

    @pytest.mark.unit
    def test_get_name_and_index_custom(self):
        p = DummyMeta()
        p.default_units = "foo"

        pset = PropertySetBase(parent=p)
        pset._defined_properties.append("foo")
        pset._defined_indices.append("bar")

        n, i = pset.get_name_and_index("foo_bar")
        assert n == "foo"
        assert i == "bar"


class DerivedPropertySet(PropertySetBase):
    foo = PropertyMetadata(name="foo")
    bar = PropertyMetadata(name="bar")
    baz = "baz"


class TestDerivedPropertySet:
    @pytest.mark.unit
    def test_standard_properties(self):
        p = DummyMeta()
        p.default_units = "foo"

        pset = DerivedPropertySet(parent=p)

        for i in pset._defined_properties:
            assert i in ["foo", "bar"]

        assert isinstance(pset.foo, PropertyMetadata)
        assert isinstance(pset.bar, PropertyMetadata)
        assert not isinstance(pset.baz, PropertyMetadata)

        # Check that we didn't accidentally overwrite anything
        assert pset.foo is not DerivedPropertySet.foo
        assert pset.bar is not DerivedPropertySet.bar
        assert pset.baz is DerivedPropertySet.baz


class TestStandardPropertySet:
    @pytest.mark.unit
    def test_defined_indices(self):
        p = DummyMeta()
        p.default_units = UnitSet()

        pset = StandardPropertySet(parent=p)

        assert pset._defined_indices == ["comp", "phase", "phase_comp"]

    @pytest.mark.unit
    def test_unindexed_properties(self):
        p = DummyMeta()
        p.default_units = UnitSet()

        pset = StandardPropertySet(parent=p)

        assert pset.temperature._indices == ["none"]
        assert pset.temperature_bubble._indices == ["none"]
        assert pset.temperature_dew._indices == ["none"]
        assert pset.temperature_red._indices == ["none"]

    @pytest.mark.unit
    def test_get_name_and_index(self):
        p = DummyMeta()
        p.default_units = UnitSet()

        pset = PropertySetBase(parent=p)
        pset._defined_properties.append("foo_bar")

        n, i = pset.get_name_and_index("foo_bar")
        assert n == "foo_bar"
        assert i is None

        n, i = pset.get_name_and_index("foo_bar_phase")
        assert n == "foo_bar"
        assert i == "phase"

        n, i = pset.get_name_and_index("foo_bar_comp")
        assert n == "foo_bar"
        assert i == "comp"

        n, i = pset.get_name_and_index("foo_bar_phase_comp")
        assert n == "foo_bar"
        assert i == "phase_comp"

        with pytest.raises(
            ValueError,
            match="Unhandled property: baz. This is mostly likely due to the "
            "property not being defined in this PropertySet.",
        ):
            pset.get_name_and_index("baz")


class TestElectrolytePropertySet:
    @pytest.mark.unit
    def test_defined_indices(self):
        p = DummyMeta()
        p.default_units = UnitSet()

        pset = ElectrolytePropertySet(parent=p)

        assert pset._defined_indices == [
            "comp",
            "phase",
            "phase_comp",
            "phase_comp_apparent",
            "phase_comp_true",
        ]

    @pytest.mark.unit
    def test_get_name_and_index(self):
        p = DummyMeta()
        p.default_units = UnitSet()

        pset = ElectrolytePropertySet(parent=p)
        pset._defined_properties.append("foo_bar")

        n, i = pset.get_name_and_index("foo_bar")
        assert n == "foo_bar"
        assert i is None

        n, i = pset.get_name_and_index("foo_bar_phase")
        assert n == "foo_bar"
        assert i == "phase"

        n, i = pset.get_name_and_index("foo_bar_comp")
        assert n == "foo_bar"
        assert i == "comp"

        n, i = pset.get_name_and_index("foo_bar_phase_comp")
        assert n == "foo_bar"
        assert i == "phase_comp"

        n, i = pset.get_name_and_index("foo_bar_phase_comp_apparent")
        assert n == "foo_bar"
        assert i == "phase_comp_apparent"

        n, i = pset.get_name_and_index("foo_bar_phase_comp_true")
        assert n == "foo_bar"
        assert i == "phase_comp_true"

        with pytest.raises(
            ValueError,
            match="Unhandled property: baz. This is mostly likely due to the "
            "property not being defined in this PropertySet.",
        ):
            pset.get_name_and_index("baz")



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Tests for flowsheet_model.

Author: Andrew Lee
"""
import pytest
import types

from pyomo.environ import ConcreteModel, Constraint, Set, Var, units as pyunits
from pyomo.common.config import ConfigBlock
from pyomo.network import Port

from idaes.core import (
    declare_process_block_class,
    PhysicalParameterBlock,
    StateBlock,
    StateBlockData,
    Phase,
    Component,
)
from idaes.core.util.exceptions import PropertyPackageError, PropertyNotSupportedError
from idaes.core.base.property_meta import PropertyClassMetadata


# -----------------------------------------------------------------------------
# Test ParameterBlock
@declare_process_block_class("ParameterBlock")
class _ParameterBlock(PhysicalParameterBlock):
    pass


@pytest.mark.unit
def test_config_block():
    # Test that PhysicalParameterBlock gets module information
    m = ConcreteModel()
    m.p = ParameterBlock()

    assert len(m.p.config) == 1
    assert isinstance(m.p.config.default_arguments, ConfigBlock)
    assert len(m.p.config.default_arguments) == 0


@pytest.mark.unit
def test_PhysicalParameterBlock():
    # Test that PhysicalParameterBlock builds correctly
    m = ConcreteModel()
    m.p = ParameterBlock()
    super(_ParameterBlock, m.p).build()


@pytest.mark.unit
def test_PhysicalParameter_NotImplementedErrors():
    # Test that class methods return NotImplementedError
    m = ConcreteModel()
    m.p = ParameterBlock()

    with pytest.raises(NotImplementedError):
        m.p.get_metadata()


@pytest.mark.unit
def test_get_phase_component_set():
    m = ConcreteModel()
    m.p = ParameterBlock()

    m.meta_object = PropertyClassMetadata()
    m.meta_object.add_default_units(
        {
            "time": pyunits.s,
            "length": pyunits.m,
            "mass": pyunits.kg,
            "amount": pyunits.mol,
            "temperature": pyunits.K,
        }
    )

    def get_metadata(self):
        return m.meta_object

    m.p.get_metadata = types.MethodType(get_metadata, m.p)

    m.p.p1 = Phase()
    m.p.p2 = Phase()
    m.p.p3 = Phase()
    m.p.a = Component()
    m.p.b = Component()
    m.p.c = Component()

    pc_set = m.p.get_phase_component_set()

    assert isinstance(m.p._phase_component_set, Set)
    assert len(m.p._phase_component_set) == 9
    for v in m.p._phase_component_set:
        assert v[0] in m.p.phase_list
        assert v[1] in m.p.component_list

    assert pc_set is m.p._phase_component_set

    # Check that method returns existing component
    # Delete phase list so that build will fail to make sure it isn't rebuilding
    m.p.del_component(m.p.phase_list)

    assert m.p.get_phase_component_set() is m.p._phase_component_set


@pytest.mark.unit
def test_get_phase_component_set_subset():
    m = ConcreteModel()
    m.p = ParameterBlock()

    m.meta_object = PropertyClassMetadata()
    m.meta_object.add_default_units(
        {
            "time": pyunits.s,
            "length": pyunits.m,
            "mass": pyunits.kg,
            "amount": pyunits.mol,
            "temperature": pyunits.K,
        }
    )

    def get_metadata(self):
        return m.meta_object

    m.p.get_metadata = types.MethodType(get_metadata, m.p)

    m.p.p1 = Phase()
    m.p.p2 = Phase(component_list=["a", "b"])
    m.p.p3 = Phase(component_list=["c"])
    m.p.a = Component()
    m.p.b = Component()
    m.p.c = Component()

    phase_comp = {"p1": ["a", "b", "c"], "p2": ["a", "b"], "p3": ["c"]}

    pc_set = m.p.get_phase_component_set()

    assert isinstance(m.p._phase_component_set, Set)
    assert len(m.p._phase_component_set) == 6
    for v in m.p._phase_component_set:
        assert v[0] in phase_comp.keys()
        assert v[1] in phase_comp[v[0]]

    assert pc_set is m.p._phase_component_set


@pytest.mark.unit
def test_get_component():
    m = ConcreteModel()
    m.p = ParameterBlock()

    m.meta_object = PropertyClassMetadata()
    m.meta_object.add_default_units(
        {
            "time": pyunits.s,
            "length": pyunits.m,
            "mass": pyunits.kg,
            "amount": pyunits.mol,
            "temperature": pyunits.K,
        }
    )

    def get_metadata(self):
        return m.meta_object

    m.p.get_metadata = types.MethodType(get_metadata, m.p)

    with pytest.raises(AttributeError):
        m.p.get_component("foo")

    m.p.comp = Component()

    assert m.p.get_component("comp") is m.p.comp

    m.p.a = object()

    with pytest.raises(
        PropertyPackageError,
        match="p get_component found an attribute a, but it does not "
        "appear to be an instance of a Component object.",
    ):
        m.p.get_component("a")


@pytest.mark.unit
def test_get_phase():
    m = ConcreteModel()
    m.p = ParameterBlock()

    with pytest.raises(AttributeError):
        m.p.get_phase("foo")

    m.p.phase = Phase()

    assert m.p.get_phase("phase") is m.p.phase

    m.p.a = object()

    with pytest.raises(
        PropertyPackageError,
        match="p get_phase found an attribute a, but it does not "
        "appear to be an instance of a Phase object.",
    ):
        m.p.get_phase("a")


@pytest.mark.unit
def test_has_inherent_reactions():
    m = ConcreteModel()
    m.p = ParameterBlock()

    assert not m.p.has_inherent_reactions

    m.p._has_inherent_reactions = True

    assert m.p.has_inherent_reactions


# -----------------------------------------------------------------------------
# Test StateBlock
@declare_process_block_class("DummyStateBlock", block_class=StateBlock)
class _StateBlockData(StateBlockData):
    def build(self):
        super(StateBlockData, self).build()


@declare_process_block_class("DummyStateBlock2", block_class=StateBlock)
class _StateBlockData2(StateBlockData):
    def build(self):
        super(StateBlockData, self).build()

    def define_state_vars(self):
        return {}


@pytest.mark.unit
def test_StateBlockBase_initialize():
    # Test that StateBlock initialize method raises NotImplementedError
    m = ConcreteModel()
    # Need to index block so that it does not do multiple inheritance
    m.s = Set(initialize=[1, 2])
    m.p = DummyStateBlock2(m.s)

    with pytest.raises(NotImplementedError):
        m.p.initialize()


@pytest.mark.unit
def test_StateBlockBase_report():
    # Test that StateBlock initialize method raises NotImplementedError
    m = ConcreteModel()
    # Need to index block so that it does not do multiple inheritance
    m.s = Set(initialize=[0, 1])
    m.p = DummyStateBlock2(m.s)

    m.p.report(dof=True)


# -----------------------------------------------------------------------------
# Test StateBlockData
@pytest.mark.unit
def test_StateBlock_config():
    # Test that StateBlockData config has correct arguments
    m = ConcreteModel()
    m.p = DummyStateBlock()

    assert len(m.p.config) == 3
    assert hasattr(m.p.config, "has_phase_equilibrium")
    assert hasattr(m.p.config, "defined_state")
    assert hasattr(m.p.config, "parameters")

    m.p.config.has_phase_equilibrium = True
    m.p.config.has_phase_equilibrium = False
    with pytest.raises(ValueError):
        m.p.config.has_phase_equilibrium = "foo"
    with pytest.raises(ValueError):
        m.p.config.has_phase_equilibrium = 10

    m.p.config.defined_state = True
    m.p.config.defined_state = False
    with pytest.raises(ValueError):
        m.p.config.defined_state = "foo"
    with pytest.raises(ValueError):
        m.p.config.defined_state = 10


@pytest.mark.unit
def test_StateBlock_NotImplementedErrors():
    # Test that placeholder methods return NotImplementedErrors
    m = ConcreteModel()
    m.p = DummyStateBlock()

    with pytest.raises(NotImplementedError):
        m.p.define_state_vars()
    with pytest.raises(NotImplementedError):
        m.p.define_port_members()
    with pytest.raises(NotImplementedError):
        m.p.get_material_flow_terms()
    with pytest.raises(NotImplementedError):
        m.p.get_material_density_terms()
    with pytest.raises(NotImplementedError):
        m.p.get_material_diffusion_terms()
    with pytest.raises(NotImplementedError):
        m.p.get_enthalpy_flow_terms()
    with pytest.raises(NotImplementedError):
        m.p.get_energy_density_terms()
    with pytest.raises(NotImplementedError):
        m.p.get_energy_diffusion_terms()
    with pytest.raises(NotImplementedError):
        m.p.calculate_bubble_point_temperature()
    with pytest.raises(NotImplementedError):
        m.p.calculate_dew_point_temperature()
    with pytest.raises(NotImplementedError):
        m.p.calculate_bubble_point_pressure()
    with pytest.raises(NotImplementedError):
        m.p.calculate_dew_point_pressure()


@pytest.mark.unit
def test_StateBlock_build_port_1index():
    m = ConcreteModel()

    m.state_block = DummyStateBlock([1, 2, 3])

    # Need to add define_port_members method to all state blocks
    def define_port_members(blk):
        return {
            "ScalarVar": blk.scalar_var,
            "IndexedVar": blk.indexed_var,
        }

    for sbd in m.state_block.values():
        # Add a scalar var to all state blocks
        sbd.scalar_var = Var(initialize=1)

        # Add an indexed block to all state blocks
        sbd.indexed_var = Var([5, 6], initialize=42)

        # Set define_port_members method
        sbd.define_port_members = types.MethodType(define_port_members, sbd)

    # Call build_port
    port, ref_name_list = m.state_block.build_port("test_doc")

    # Check Port and members
    assert isinstance(port, Port)
    assert port.doc == "test_doc"

    for i in m.state_block:
        assert port.ScalarVar[i] is m.state_block[i].scalar_var
        for j in m.state_block[i].indexed_var:
            assert port.IndexedVar[i, j] is m.state_block[i].indexed_var[j]

    # Check References
    assert len(ref_name_list) == 2
    for ref, cname in ref_name_list:
        assert cname in ["ScalarVar", "IndexedVar"]
        if cname == "ScalarVar":
            assert isinstance(ref, Var)
            assert ref is port.ScalarVar
            for i, v in ref.items():
                assert i in [1, 2, 3]
                assert v is m.state_block[i].scalar_var
        elif cname == "IndexedVar":
            assert isinstance(ref, Var)
            assert ref is port.IndexedVar
            for i, v in ref.items():
                assert i[0] in [1, 2, 3]
                assert i[1] in [5, 6]
                assert v is m.state_block[i[0]].indexed_var[i[1]]
        else:
            # Catch for unexpected name
            raise ValueError


@pytest.mark.unit
def test_StateBlock_build_port_2index():
    m = ConcreteModel()

    m.state_block = DummyStateBlock([1, 2, 3], [10, 20])

    # Need to add define_port_members method to all state blocks
    def define_port_members(blk):
        return {
            "ScalarVar": blk.scalar_var,
            "IndexedVar": blk.indexed_var,
        }

    for sbd in m.state_block.values():
        # Add a scalar var to all state blocks
        sbd.scalar_var = Var(initialize=1)

        # Add an indexed block to all state blocks
        sbd.indexed_var = Var([5, 6], initialize=42)

        # Set define_port_members method
        sbd.define_port_members = types.MethodType(define_port_members, sbd)

    # Call build_port
    port, ref_name_list = m.state_block.build_port("test_doc")

    # Check Port and members
    assert isinstance(port, Port)
    assert port.doc == "test_doc"

    for i in m.state_block:
        assert port.ScalarVar[i] is m.state_block[i].scalar_var
        for j in m.state_block[i].indexed_var:
            assert port.IndexedVar[i, j] is m.state_block[i].indexed_var[j]

    # Check References
    assert len(ref_name_list) == 2
    for ref, cname in ref_name_list:
        assert cname in ["ScalarVar", "IndexedVar"]
        if cname == "ScalarVar":
            assert isinstance(ref, Var)
            assert ref is port.ScalarVar
            for i, v in ref.items():
                assert i[0] in [1, 2, 3]
                assert i[1] in [10, 20]
                assert v is m.state_block[i].scalar_var
        elif cname == "IndexedVar":
            assert isinstance(ref, Var)
            assert ref is port.IndexedVar
            for i, v in ref.items():
                assert i[0] in [1, 2, 3]
                assert i[1] in [10, 20]
                assert i[2] in [5, 6]
                assert v is m.state_block[i[0], i[1]].indexed_var[i[2]]
        else:
            # Catch for unexpected name
            raise ValueError


@pytest.mark.unit
def test_StateBlock_build_port_2index_subset():
    m = ConcreteModel()

    m.state_block = DummyStateBlock([1, 2, 3], [10, 20])

    # Need to add define_port_members method to all state blocks
    def define_port_members(blk):
        return {
            "ScalarVar": blk.scalar_var,
            "IndexedVar": blk.indexed_var,
        }

    for sbd in m.state_block.values():
        # add a scalar var to all state blocks
        sbd.scalar_var = Var(initialize=1)

        # Add an indexed block to all state blocks
        sbd.indexed_var = Var([5, 6], initialize=42)

        # Set define_port_members method
        sbd.define_port_members = types.MethodType(define_port_members, sbd)

    # Call build_port
    port, ref_name_list = m.state_block.build_port("test_doc")

    # Check Port and members
    assert isinstance(port, Port)
    assert port.doc == "test_doc"

    for i in m.state_block:
        assert port.ScalarVar[i] is m.state_block[i].scalar_var
        for j in m.state_block[i].indexed_var:
            assert port.IndexedVar[i, j] is m.state_block[i].indexed_var[j]

    # Check References
    assert len(ref_name_list) == 2
    for ref, cname in ref_name_list:
        assert cname in ["ScalarVar", "IndexedVar"]
        if cname == "ScalarVar":
            assert isinstance(ref, Var)
            assert ref is port.ScalarVar
            for i, v in ref.items():
                assert i[0] in [1, 2, 3]
                assert i[1] in [10, 20]
                assert v is m.state_block[i].scalar_var
        elif cname == "IndexedVar":
            assert isinstance(ref, Var)
            assert ref is port.IndexedVar
            for i, v in ref.items():
                assert i[0] in [1, 2, 3]
                assert i[1] in [10, 20]
                assert i[2] in [5, 6]
                assert v is m.state_block[i[0], i[1]].indexed_var[i[2]]
        else:
            # Catch for unexpected name
            raise ValueError


# -----------------------------------------------------------------------------
# Test parameter block reference attribute
@declare_process_block_class("Parameters")
class _Parameters(PhysicalParameterBlock):
    def build(self):
        super(_Parameters, self).build()

        self.p1 = Phase()
        self.c1 = Component()

    @classmethod
    def define_metadata(cls, obj):
        obj.define_custom_properties(
            {
                "a": {"method": "a_method"},
                "recursion1": {"method": "_recursion1"},
                "recursion2": {"method": "_recursion2"},
                "not_callable": {"method": "test_obj"},
                "raise_exception": {"method": "_raise_exception"},
                "not_supported": {"supported": False},
                "does_not_create_component": {"method": "_does_not_create_component"},
            }
        )
        obj.add_default_units(
            {
                "time": pyunits.s,
                "mass": pyunits.kg,
                "length": pyunits.m,
                "amount": pyunits.mol,
                "temperature": pyunits.K,
            }
        )


@declare_process_block_class("StateTest", block_class=StateBlock)
class _StateTest(StateBlockData):
    def build(self):
        super(_StateTest, self).build()


@pytest.mark.unit
def test_param_ref():
    m = ConcreteModel()
    m.pb = Parameters()
    m.p = StateTest(parameters=m.pb)

    assert m.p.params == m.p.config.parameters


@pytest.mark.unit
def test_validate_params():
    # Test that validate params has been triggered
    m = ConcreteModel()
    m.pb = Parameters()
    m.p = StateTest(parameters=m.pb)

    # If validation has been triggered, Phase & Component objects should exist
    assert isinstance(m.pb.p1, Phase)
    assert isinstance(m.pb.c1, Component)


@pytest.mark.unit
def test_has_inherent_reactions_state_block():
    m = ConcreteModel()
    m.pb = Parameters()
    m.p = StateTest(parameters=m.pb)

    assert not m.p.has_inherent_reactions

    m.pb._has_inherent_reactions = True

    assert m.p.has_inherent_reactions


# -----------------------------------------------------------------------------
# Test properties __getattr__ method
@declare_process_block_class("State", block_class=StateBlock)
class _State(StateBlockData):
    def build(self):
        super(StateBlockData, self).build()

        self.test_obj = 1

    def a_method(self):
        self.a = Var(initialize=1)

    def _recursion1(self):
        self.recursive_cons1 = Constraint(expr=self.recursion2 == 1)

    def _recursion2(self):
        self.recursive_cons2 = Constraint(expr=self.recursion1 == 1)

    def _raise_exception(self):
        # PYLINT-TODO
        # pylint: disable-next=broad-exception-raised
        raise Exception()

    def _does_not_create_component(self):
        pass


@pytest.fixture()
def m():
    m = ConcreteModel()
    m.pb = Parameters()
    m.p = State(parameters=m.pb)

    return m


@pytest.mark.unit
def test_getattr_add_var(m):
    assert isinstance(m.p.a, Var)
    assert m.p.a.value == 1
    assert m.p.is_property_constructed("a")


@pytest.mark.unit
def test_lock_attributes(m):
    with pytest.raises(AttributeError):
        with m.p.lock_attribute_creation_context():
            m.p.a.value == 1
    # Make sure it unlocked
    assert m.p.a.value == 1


@pytest.mark.unit
def test_is_property_constructed(m):
    assert not m.p.is_property_constructed("a")
    assert m.p.a.value == 1
    assert m.p.is_property_constructed("a")


@pytest.mark.unit
def test_getattr_protected(m):
    with pytest.raises(PropertyPackageError):
        # Call a protected component that does not exist
        m.p.cons = Constraint(expr=m.p._foo == 1)


@pytest.mark.unit
def test_getattr_recursion(m):
    with pytest.raises(PropertyPackageError):
        # Call a component that triggers a recursive loop of calls
        m.p.cons = Constraint(expr=m.p.recursion1 == 1)


@pytest.mark.unit
def test_getattr_does_not_exist(m):
    with pytest.raises(PropertyNotSupportedError):
        m.p.cons = Constraint(expr=m.p.does_not_exist == 1)


@pytest.mark.unit
def test_getattr_not_callable(m):
    with pytest.raises(PropertyPackageError):
        m.p.cons = Constraint(expr=m.p.not_callable == 1)


@pytest.mark.unit
def test_getattr_not_supported(m):
    with pytest.raises(PropertyNotSupportedError):
        m.p.cons = Constraint(expr=m.p.not_supported == 1)


@pytest.mark.unit
def test_getattr_raise_exception(m):
    with pytest.raises(Exception):
        m.p.cons = Constraint(expr=m.p.raise_exception == 1)


# TODO : Need a test for cases where method does not create property
# @pytest.mark.unit
# def test_getattr_does_not_create_component(m):
#    with pytest.raises(PropertyPackageError):
#        m.p.cons = Constraint(expr=m.p.does_not_create_component == 1)



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
1-D Cross Flow Heat Exchanger Model With Wall Temperatures

Discretization based on tube rows
"""


# Import Pyomo libraries
from pyomo.environ import (
    assert_optimal_termination,
    Block,
    value,
    Var,
    log,
    Param,
    Reference,
    units as pyunits,
)
from pyomo.common.config import ConfigValue, In, Bool
from pyomo.network import Port

# Import IDAES cores
from idaes.core import declare_process_block_class
from idaes.core.util.constants import Constants as const
import idaes.core.util.scaling as iscale
from idaes.core.solvers import get_solver
from idaes.core.util.exceptions import InitializationError
from idaes.core.util.misc import add_object_reference
from idaes.core.util.exceptions import ConfigurationError, BurntToast
import idaes.logger as idaeslog
from idaes.core.util.tables import create_stream_table_dataframe
from idaes.models.unit_models.heat_exchanger import (
    HeatExchangerFlowPattern,
)
from idaes.models.unit_models.heat_exchanger_1D import HeatExchanger1DData
from idaes.models_extra.power_generation.unit_models import heat_exchanger_common
from idaes.core.initialization import SingleControlVolumeUnitInitializer


__author__ = "Jinliang Ma, Douglas Allan"


class CrossFlowHeatExchanger1DInitializer(SingleControlVolumeUnitInitializer):
    """
    Initializer for Cross Flow Heat Exchanger 1D units.

    First, the shell and tube control volumes are initialized without heat transfer. Next
    the total possible heat transfer between streams is estimated based on heat capacity,
    flow rate, and inlet/outlet temperatures. The actual temperature change is set to be
    half the theoretical maximum, and the shell and tube are initialized with linear
    temperature profiles. Finally, temperatures besides the inlets are unfixed and
    the performance equations are activated before a full solve of the system model.
    """

    def initialize_main_model(
        self,
        model: Block,
        copy_inlet_state: bool = False,
    ):
        """
        Initialization routine for the main Cross Flow Heat Exchanger 1D model
        (as opposed to submodels like costing, which presently do not exist).

        Args:
            model: Pyomo Block to be initialized.
            copy_inlet_state: bool (default=False). Whether to copy inlet state to other states or not
                (0-D control volumes only). Copying will generally be faster, but inlet states may not contain
                all properties required elsewhere.
            duty: initial guess for heat duty to assist with initialization. Can be a Pyomo expression with units.

        Returns:
            Pyomo solver results object.

        """
        # Set solver options
        init_log = idaeslog.getInitLogger(
            model.name, self.get_output_level(), tag="unit"
        )
        solve_log = idaeslog.getSolveLogger(
            model.name, self.get_output_level(), tag="unit"
        )

        solver_obj = get_solver(self.config.solver, self.config.solver_options)

        hot_side = model.hot_side
        cold_side = model.cold_side
        t0 = model.flowsheet().time.first()
        if not (
            "temperature" in hot_side.properties[t0, 0].define_state_vars().keys()
            and "temperature" in cold_side.properties[t0, 0].define_state_vars().keys()
        ):
            raise NotImplementedError(
                "Presently, initialization of the CrossFlowHeatExchanger1D requires "
                "temperature to be a state variable of both hot side and cold side "
                "property packages. Extension to enth_mol or enth_mass as state variables "
                "is straightforward---feel free to open a pull request implementing it."
            )

        hot_units = model.config.hot_side.property_package.get_metadata().derived_units
        cold_units = (
            model.config.cold_side.property_package.get_metadata().derived_units
        )

        if model.config.shell_is_hot:
            shell = model.hot_side
            tube = model.cold_side
            shell_has_pressure_change = model.config.hot_side.has_pressure_change
            tube_has_pressure_change = model.config.cold_side.has_pressure_change
            shell_units = (
                model.config.hot_side.property_package.get_metadata().derived_units
            )
            tube_units = (
                model.config.cold_side.property_package.get_metadata().derived_units
            )
        else:
            shell = model.cold_side
            tube = model.hot_side
            shell_has_pressure_change = model.config.cold_side.has_pressure_change
            tube_has_pressure_change = model.config.hot_side.has_pressure_change
            shell_units = (
                model.config.cold_side.property_package.get_metadata().derived_units
            )
            tube_units = (
                model.config.hot_side.property_package.get_metadata().derived_units
            )

        # Trigger creation of cp for use in future initialization
        # Important to do before initializing property packages in
        # case it is implemented as Var-Constraint pair instead of
        # an Expression
        hot_side.properties[t0, 0].cp_mol  # pylint: disable=pointless-statement
        cold_side.properties[t0, 0].cp_mol  # pylint: disable=pointless-statement

        # ---------------------------------------------------------------------
        # Initialize shell block
        self.initialize_control_volume(tube, copy_inlet_state)
        self.initialize_control_volume(shell, copy_inlet_state)

        init_log.info_high("Initialization Step 1 Complete.")

        # Set tube thermal conductivity to a small value to avoid IPOPT unable to solve initially
        therm_cond_wall_save = model.therm_cond_wall.value
        model.therm_cond_wall.set_value(0.05)
        # In Step 2, fix tube metal temperatures fix fluid state variables (enthalpy/temperature and pressure)
        # calculate maximum heat duty assuming infinite area and use half of the maximum duty as initial guess to calculate outlet temperature

        for t in model.flowsheet().config.time:
            mcp_hot_side = value(
                pyunits.convert(
                    hot_side.properties[t, 0].flow_mol
                    * hot_side.properties[t, 0].cp_mol,
                    to_units=shell_units["power"] / shell_units["temperature"],
                )
            )
            T_in_hot_side = value(
                pyunits.convert(
                    hot_side.properties[t, 0].temperature,
                    to_units=shell_units["temperature"],
                )
            )
            P_in_hot_side = value(hot_side.properties[t, 0].pressure)
            if model.config.flow_type == HeatExchangerFlowPattern.cocurrent:
                mcp_cold_side = value(
                    pyunits.convert(
                        cold_side.properties[t, 0].flow_mol
                        * cold_side.properties[t, 0].cp_mol,
                        to_units=shell_units["power"] / shell_units["temperature"],
                    )
                )
                T_in_cold_side = value(
                    pyunits.convert(
                        cold_side.properties[t, 0].temperature,
                        to_units=shell_units["temperature"],
                    )
                )
                P_in_cold_side = value(cold_side.properties[t, 0].pressure)

                T_out_max = (
                    mcp_cold_side * T_in_cold_side + mcp_hot_side * T_in_hot_side
                ) / (mcp_cold_side + mcp_hot_side)

                q_guess = mcp_cold_side * (T_out_max - T_in_cold_side) / 2

                temp_out_cold_side_guess = T_in_cold_side + q_guess / mcp_cold_side

                cold_side.properties[t, 1].temperature.fix(
                    pyunits.convert_value(
                        temp_out_cold_side_guess,
                        from_units=shell_units["temperature"],
                        to_units=cold_units["temperature"],
                    )
                )

                temp_out_hot_side_guess = T_in_cold_side - q_guess / mcp_hot_side
                hot_side.properties[t, 1].temperature.fix(
                    pyunits.convert_value(
                        temp_out_hot_side_guess,
                        from_units=shell_units["temperature"],
                        to_units=hot_units["temperature"],
                    )
                )

            elif model.config.flow_type == HeatExchangerFlowPattern.countercurrent:
                mcp_cold_side = value(
                    pyunits.convert(
                        cold_side.properties[t, 1].flow_mol
                        * cold_side.properties[t, 1].cp_mol,
                        to_units=shell_units["power"] / shell_units["temperature"],
                    )
                )
                T_in_cold_side = value(
                    pyunits.convert(
                        cold_side.properties[t, 1].temperature,
                        to_units=shell_units["temperature"],
                    )
                )
                P_in_cold_side = value(cold_side.properties[t, 1].pressure)

                if mcp_cold_side < mcp_hot_side:
                    q_guess = mcp_cold_side * (T_in_hot_side - T_in_cold_side) / 2
                else:
                    q_guess = mcp_hot_side * (T_in_hot_side - T_in_cold_side) / 2

                temp_out_cold_side_guess = T_in_cold_side + q_guess / mcp_cold_side
                cold_side.properties[t, 0].temperature.fix(
                    pyunits.convert_value(
                        temp_out_cold_side_guess,
                        from_units=shell_units["temperature"],
                        to_units=cold_units["temperature"],
                    )
                )

                temp_out_hot_side_guess = T_in_hot_side - q_guess / mcp_hot_side
                hot_side.properties[t, 1].temperature.fix(
                    pyunits.convert_value(
                        temp_out_hot_side_guess,
                        from_units=shell_units["temperature"],
                        to_units=hot_units["temperature"],
                    )
                )

            else:
                raise BurntToast(
                    "HeatExchangerFlowPattern should be limited to cocurrent "
                    "or countercurrent flow by parent model. Please open an "
                    "issue on the IDAES Github so this error can be fixed."
                )

            for z in cold_side.length_domain:
                hot_side.properties[t, z].temperature.fix(
                    value(
                        (1 - z) * hot_side.properties[t, 0].temperature
                        + z * hot_side.properties[t, 1].temperature
                    )
                )
                cold_side.properties[t, z].temperature.fix(
                    value(
                        (1 - z) * cold_side.properties[t, 0].temperature
                        + z * cold_side.properties[t, 1].temperature
                    )
                )
                model.temp_wall_center[t, z].fix(
                    value(
                        pyunits.convert(
                            hot_side.properties[t, z].temperature,
                            to_units=shell_units["temperature"],
                        )
                        + pyunits.convert(
                            cold_side.properties[t, z].temperature,
                            to_units=shell_units["temperature"],
                        )
                    )
                    / 2
                )

                model.temp_wall_shell[t, z].set_value(
                    model.temp_wall_center[t, z].value
                )
                model.temp_wall_tube[t, z].set_value(
                    pyunits.convert_value(
                        model.temp_wall_center[t, z].value,
                        from_units=shell_units["temperature"],
                        to_units=tube_units["temperature"],
                    )
                )

                if model.config.cold_side.has_pressure_change:
                    cold_side.properties[t, z].pressure.fix(P_in_cold_side)
                if model.config.hot_side.has_pressure_change:
                    hot_side.properties[t, z].pressure.fix(P_in_hot_side)

        model.temp_wall_center_eqn.deactivate()
        if tube_has_pressure_change:
            model.deltaP_tube_eqn.deactivate()
        if shell_has_pressure_change:
            model.deltaP_shell_eqn.deactivate()
        model.heat_tube_eqn.deactivate()
        model.heat_shell_eqn.deactivate()

        with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
            res = solver_obj.solve(model, tee=slc.tee)
        try:
            assert_optimal_termination(res)
        except AssertionError:
            raise InitializationError("Initialization solve failed.")
        init_log.info_high("Initialization Step 2 {}.".format(idaeslog.condition(res)))

        # In Step 3, unfix fluid state variables (enthalpy/temperature and pressure)
        # keep the inlet state variables fixed, otherwise, the degree of freedom > 0
        hot_side.properties[:, :].temperature.unfix()
        hot_side.properties[:, :].pressure.unfix()
        hot_side.properties[:, 0].temperature.fix()
        hot_side.properties[:, 0].pressure.fix()

        cold_side.properties[:, :].temperature.unfix()
        cold_side.properties[:, :].pressure.unfix()
        if model.config.flow_type == HeatExchangerFlowPattern.cocurrent:
            cold_side.properties[:, 0].temperature.fix()
            cold_side.properties[:, 0].pressure.fix()
        elif model.config.flow_type == HeatExchangerFlowPattern.countercurrent:
            cold_side.properties[:, 1].temperature.fix()
            cold_side.properties[:, 1].pressure.fix()
        else:
            raise BurntToast(
                "HeatExchangerFlowPattern should be limited to cocurrent "
                "or countercurrent flow by parent model. Please open an "
                "issue on the IDAES Github so this error can be fixed."
            )

        if tube_has_pressure_change:
            model.deltaP_tube_eqn.activate()
        if shell_has_pressure_change:
            model.deltaP_shell_eqn.activate()
        model.heat_tube_eqn.activate()
        model.heat_shell_eqn.activate()

        with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
            res = solver_obj.solve(model, tee=slc.tee)
        try:
            assert_optimal_termination(res)
        except AssertionError:
            raise InitializationError("Initialization solve failed.")

        init_log.info_high("Initialization Step 3 {}.".format(idaeslog.condition(res)))

        model.temp_wall_center.unfix()
        model.temp_wall_center_eqn.activate()

        with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
            res = solver_obj.solve(model, tee=slc.tee)
        try:
            assert_optimal_termination(res)
        except AssertionError:
            raise InitializationError("Initialization solve failed.")

        init_log.info_high("Initialization Step 4 {}.".format(idaeslog.condition(res)))

        # set the wall thermal conductivity back to the user specified value
        model.therm_cond_wall.set_value(therm_cond_wall_save)

        with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
            res = solver_obj.solve(model, tee=slc.tee)
        init_log.info_high("Initialization Step 5 {}.".format(idaeslog.condition(res)))
        init_log.info("Initialization Complete.")
        return res


@declare_process_block_class("CrossFlowHeatExchanger1D")
class CrossFlowHeatExchanger1DData(HeatExchanger1DData):
    """Standard Cross Flow Heat Exchanger Unit Model Class."""

    default_initializer = CrossFlowHeatExchanger1DInitializer

    CONFIG = HeatExchanger1DData.CONFIG()
    CONFIG.declare(
        "shell_is_hot",
        ConfigValue(
            default=True,
            domain=Bool,
            description="Shell side contains hot fluid",
            doc="""Boolean flag indicating whether shell side contains hot fluid (default=True).
        If True, shell side will be the hot_side, if False shell side will be cold_side.""",
        ),
    )
    CONFIG.declare(
        "tube_arrangement",
        ConfigValue(
            default="in-line",
            domain=In(["in-line", "staggered"]),
            description="tube configuration",
            doc="tube arrangement could be in-line or staggered",
        ),
    )

    def _process_config(self):
        # Copy and pasted from ShellAndTube1D
        super()._process_config()

        # Check for custom names, and if not present assign defaults
        if self.config.hot_side_name is None:
            if self.config.shell_is_hot:
                self.config.hot_side_name = "Shell"
            else:
                self.config.hot_side_name = "Tube"

        if self.config.cold_side_name is None:
            if self.config.shell_is_hot:
                self.config.cold_side_name = "Tube"
            else:
                self.config.cold_side_name = "Shell"

    def build(self):
        """
        Begin building model (pre-DAE transformation).

        Args:
            None

        Returns:
            None
        """
        # Call HeatExchanger1DData build to make common components
        super().build()

        # The HeatExchanger1DData equates the heat lost by the hot side and heat gained by the cold side.
        # That equation is deleted here because heat can accumulate in the wall.
        self.del_component(self.heat_conservation)

        # Create aliases for ports
        if self.config.shell_is_hot:
            self.shell_inlet = Port(extends=self.hot_side_inlet)
            self.shell_outlet = Port(extends=self.hot_side_outlet)
            self.tube_inlet = Port(extends=self.cold_side_inlet)
            self.tube_outlet = Port(extends=self.cold_side_outlet)
        else:
            self.shell_inlet = Port(extends=self.cold_side_inlet)
            self.shell_outlet = Port(extends=self.cold_side_outlet)
            self.tube_inlet = Port(extends=self.hot_side_inlet)
            self.tube_outlet = Port(extends=self.hot_side_outlet)

    def _make_geometry(self):
        """
        Constraints for unit model.

        Args:
            None

        Returns:
            None
        """
        if self.config.shell_is_hot:
            shell = self.hot_side
            tube = self.cold_side
            shell_units = (
                self.config.hot_side.property_package.get_metadata().derived_units
            )
        else:
            shell = self.cold_side
            tube = self.hot_side
            shell_units = (
                self.config.cold_side.property_package.get_metadata().derived_units
            )
        # Use add_object_reference for scalar vars
        add_object_reference(self, "area_flow_shell", shell.area)
        add_object_reference(self, "area_flow_tube", tube.area)
        add_object_reference(self, "length_flow_shell", shell.length)
        add_object_reference(self, "length_flow_tube", tube.length)

        heat_exchanger_common.make_geometry_common(self, shell_units=shell_units)

        # Important that these values about tube geometry are in shell units!
        @self.Constraint(doc="Length of tube side flow")
        def length_flow_tube_eqn(b):
            return (
                pyunits.convert(b.length_flow_tube, to_units=shell_units["length"])
                == b.number_passes * b.length_tube_seg
            )

        @self.Constraint(doc="Total area of tube flow")
        def area_flow_tube_eqn(b):
            return (
                b.area_flow_tube
                == 0.25
                * const.pi
                * b.di_tube**2.0
                * b.number_columns_per_pass
                * b.number_rows_per_pass
            )

    def _make_performance(self):
        """
        Constraints for unit model.

        Args:
            None

        Returns:
            None
        """
        if self.config.shell_is_hot:
            shell = self.hot_side
            tube = self.cold_side
            shell_units = (
                self.config.hot_side.property_package.get_metadata().derived_units
            )
            tube_units = (
                self.config.cold_side.property_package.get_metadata().derived_units
            )
        else:
            shell = self.cold_side
            tube = self.hot_side
            shell_units = (
                self.config.cold_side.property_package.get_metadata().derived_units
            )
            tube_units = (
                self.config.hot_side.property_package.get_metadata().derived_units
            )
        if (
            len(self.config.hot_side.property_package.phase_list) != 1
            or len(self.config.cold_side.property_package.phase_list) != 1
        ):
            raise ConfigurationError(
                "The CrossFlowHeatExchanger1D model is valid only for property packages "
                f"with a single phase. Found {len(self.config.hot_side.property_package.phase_list)} "
                f"phases on the hot side and {len(self.config.cold_side.property_package.phase_list)} "
                "phases on the cold side."
            )

        p_hot = self.config.hot_side.property_package.phase_list.at(1)
        pobj_hot = self.config.hot_side.property_package.get_phase(p_hot)
        p_cold = self.config.cold_side.property_package.phase_list.at(1)
        pobj_cold = self.config.cold_side.property_package.get_phase(p_cold)
        if not pobj_hot.is_vapor_phase():
            raise ConfigurationError(
                "The CrossFlowHeatExchanger1D model is valid only for property packages "
                "whose single phase is a vapor phase. The hot side phase is not a vapor phase."
            )
        if not pobj_cold.is_vapor_phase():
            raise ConfigurationError(
                "The CrossFlowHeatExchanger1D model is valid only for property packages "
                "whose single phase is a vapor phase. The cold side phase is not a vapor phase."
            )

        self.heat_tube = Reference(tube.heat)
        self.heat_shell = Reference(shell.heat)

        shell_has_pressure_change = False
        tube_has_pressure_change = False

        if self.config.cold_side.has_pressure_change:
            if self.config.shell_is_hot:
                self.deltaP_tube = Reference(tube.deltaP)
                tube_has_pressure_change = True
            else:
                self.deltaP_shell = Reference(shell.deltaP)
                shell_has_pressure_change = True
        if self.config.hot_side.has_pressure_change:
            if self.config.shell_is_hot:
                self.deltaP_shell = Reference(shell.deltaP)
                shell_has_pressure_change = True
            else:
                self.deltaP_tube = Reference(tube.deltaP)
                tube_has_pressure_change = True

        heat_exchanger_common.make_performance_common(
            self,
            shell=shell,
            shell_units=shell_units,
            shell_has_pressure_change=shell_has_pressure_change,
            make_reynolds=True,
            make_nusselt=True,
        )

        self.heat_transfer_coeff_tube = Var(
            self.flowsheet().config.time,
            tube.length_domain,
            initialize=100.0,
            units=tube_units["heat_transfer_coefficient"],
            doc="tube side convective heat transfer coefficient",
        )

        # Heat transfer resistance due to the fouling on tube side
        self.rfouling_tube = Param(
            initialize=0.0,
            mutable=True,
            units=1 / tube_units["heat_transfer_coefficient"],
            doc="fouling resistance on tube side",
        )
        # Correction factor for convective heat transfer coefficient on tube side
        self.fcorrection_htc_tube = Var(
            initialize=1.0, doc="correction factor for convective HTC on tube side"
        )
        # Correction factor for tube side pressure drop due to friction
        if tube_has_pressure_change:
            # Loss coefficient for a 180 degree bend (u-turn), usually related to radius to inside diameter ratio
            self.kloss_uturn = Param(
                initialize=0.5, mutable=True, doc="loss coefficient of a tube u-turn"
            )
            self.fcorrection_dp_tube = Var(
                initialize=1.0, doc="correction factor for tube side pressure drop"
            )

        # Boundary wall temperature on tube side
        self.temp_wall_tube = Var(
            self.flowsheet().config.time,
            tube.length_domain,
            initialize=500,
            units=tube_units[
                "temperature"
            ],  # Want to be in shell units for consistency in equations
            doc="boundary wall temperature on tube side",
        )
        # Tube side heat transfer coefficient and pressure drop
        # -----------------------------------------------------
        # Velocity on tube side
        self.v_tube = Var(
            self.flowsheet().config.time,
            tube.length_domain,
            initialize=1.0,
            units=tube_units["velocity"],
            doc="velocity on tube side",
        )

        # Reynalds number on tube side
        self.N_Re_tube = Var(
            self.flowsheet().config.time,
            tube.length_domain,
            initialize=10000.0,
            units=pyunits.dimensionless,
            doc="Reynolds number on tube side",
            bounds=(1e-7, None),
        )
        if tube_has_pressure_change:
            # Friction factor on tube side
            self.friction_factor_tube = Var(
                self.flowsheet().config.time,
                tube.length_domain,
                initialize=1.0,
                doc="friction factor on tube side",
            )

            # Pressure drop due to friction on tube side
            self.deltaP_tube_friction = Var(
                self.flowsheet().config.time,
                tube.length_domain,
                initialize=-10.0,
                units=tube_units["pressure"] / tube_units["length"],
                doc="pressure drop due to friction on tube side",
            )

            # Pressure drop due to 180 degree turn on tube side
            self.deltaP_tube_uturn = Var(
                self.flowsheet().config.time,
                tube.length_domain,
                initialize=-10.0,
                units=tube_units["pressure"] / tube_units["length"],
                doc="pressure drop due to u-turn on tube side",
            )
        # Nusselt number on tube side
        self.N_Nu_tube = Var(
            self.flowsheet().config.time,
            tube.length_domain,
            initialize=1,
            units=pyunits.dimensionless,
            doc="Nusselts number on tube side",
            bounds=(1e-7, None),
        )

        # Velocity equation
        @self.Constraint(
            self.flowsheet().config.time,
            tube.length_domain,
            doc="tube side velocity equation",
        )
        def v_tube_eqn(b, t, x):
            return (
                b.v_tube[t, x]
                * pyunits.convert(b.area_flow_tube, to_units=tube_units["area"])
                * tube.properties[t, x].dens_mol_phase["Vap"]
                == tube.properties[t, x].flow_mol
            )

        # Reynolds number
        @self.Constraint(
            self.flowsheet().config.time,
            tube.length_domain,
            doc="Reynolds number equation on tube side",
        )
        def N_Re_tube_eqn(b, t, x):
            return (
                b.N_Re_tube[t, x] * tube.properties[t, x].visc_d_phase["Vap"]
                == pyunits.convert(b.di_tube, to_units=tube_units["length"])
                * b.v_tube[t, x]
                * tube.properties[t, x].dens_mol_phase["Vap"]
                * tube.properties[t, x].mw
            )

        if tube_has_pressure_change:
            # Friction factor
            @self.Constraint(
                self.flowsheet().config.time,
                tube.length_domain,
                doc="Darcy friction factor on tube side",
            )
            def friction_factor_tube_eqn(b, t, x):
                return (
                    b.friction_factor_tube[t, x] * b.N_Re_tube[t, x] ** 0.25
                    == 0.3164 * b.fcorrection_dp_tube
                )

            # Pressure drop due to friction per tube length
            @self.Constraint(
                self.flowsheet().config.time,
                tube.length_domain,
                doc="pressure drop due to friction per tube length",
            )
            def deltaP_tube_friction_eqn(b, t, x):
                return (
                    b.deltaP_tube_friction[t, x]
                    * pyunits.convert(b.di_tube, to_units=tube_units["length"])
                    == -0.5
                    * tube.properties[t, x].dens_mass_phase["Vap"]
                    * b.v_tube[t, x] ** 2
                    * b.friction_factor_tube[t, x]
                )

            # Pressure drop due to u-turn
            @self.Constraint(
                self.flowsheet().config.time,
                tube.length_domain,
                doc="pressure drop due to u-turn on tube side",
            )
            def deltaP_tube_uturn_eqn(b, t, x):
                return (
                    b.deltaP_tube_uturn[t, x]
                    * pyunits.convert(b.length_tube_seg, to_units=tube_units["length"])
                    == -0.5
                    * tube.properties[t, x].dens_mass_phase["Vap"]
                    * b.v_tube[t, x] ** 2
                    * b.kloss_uturn
                )

            # Total pressure drop on tube side
            @self.Constraint(
                self.flowsheet().config.time,
                tube.length_domain,
                doc="total pressure drop on tube side",
            )
            def deltaP_tube_eqn(b, t, x):
                return b.deltaP_tube[t, x] == (
                    b.deltaP_tube_friction[t, x] + b.deltaP_tube_uturn[t, x]
                )

        @self.Constraint(
            self.flowsheet().config.time,
            tube.length_domain,
            doc="convective heat transfer coefficient equation on tube side",
        )
        def heat_transfer_coeff_tube_eqn(b, t, x):
            return (
                b.heat_transfer_coeff_tube[t, x] * b.di_tube
                == b.N_Nu_tube[t, x]
                * tube.properties[t, x].therm_cond_phase["Vap"]
                * b.fcorrection_htc_tube
            )

        # Nusselts number
        @self.Constraint(
            self.flowsheet().config.time,
            tube.length_domain,
            doc="Nusselts number equation on tube side",
        )
        def N_Nu_tube_eqn(b, t, x):
            return (
                b.N_Nu_tube[t, x]
                == 0.023
                * b.N_Re_tube[t, x] ** 0.8
                * tube.properties[t, x].prandtl_number_phase["Vap"] ** 0.4
            )

        @self.Constraint(
            self.flowsheet().config.time,
            shell.length_domain,
            doc="Nusselts number equation on shell side",
        )
        def N_Nu_shell_eqn(b, t, x):
            return (
                b.N_Nu_shell[t, x]
                == b.f_arrangement
                * 0.33
                * b.N_Re_shell[t, x] ** 0.6
                * shell.properties[t, x].prandtl_number_phase["Vap"] ** 0.333333
            )

        # Energy balance with tube wall
        # ------------------------------------
        # Heat to wall per length on tube side
        @self.Constraint(
            self.flowsheet().config.time,
            tube.length_domain,
            doc="heat per length on tube side",
        )
        def heat_tube_eqn(b, t, x):
            return b.heat_tube[t, x] == (
                b.heat_transfer_coeff_tube[t, x]
                * const.pi
                * pyunits.convert(b.di_tube, to_units=tube_units["length"])
                * b.number_rows_per_pass
                * b.number_columns_per_pass
                * (b.temp_wall_tube[t, x] - tube.properties[t, x].temperature)
            )

        # Heat to wall per length on shell side
        @self.Constraint(
            self.flowsheet().config.time,
            shell.length_domain,
            doc="heat per length on shell side",
        )
        def heat_shell_eqn(b, t, x):
            return b.heat_shell[t, x] * b.length_flow_shell == pyunits.convert(
                b.length_flow_tube, to_units=shell_units["length"]
            ) * b.total_heat_transfer_coeff_shell[
                t, x
            ] * const.pi * b.do_tube * b.number_rows_per_pass * b.number_columns_per_pass * (
                b.temp_wall_shell[t, x] - shell.properties[t, x].temperature
            )

        # Tube side wall temperature
        # FIXME replace with deviation variables
        @self.Constraint(
            self.flowsheet().config.time,
            tube.length_domain,
            doc="tube side wall temperature",
        )
        def temp_wall_tube_eqn(b, t, x):
            return b.heat_transfer_coeff_tube[t, x] * (
                tube.properties[t, x].temperature - b.temp_wall_tube[t, x]
            ) * (
                pyunits.convert(
                    b.thickness_tube / 2 / b.therm_cond_wall,
                    to_units=1 / tube_units["heat_transfer_coefficient"],
                )
                + b.rfouling_tube
            ) == b.temp_wall_tube[
                t, x
            ] - pyunits.convert(
                b.temp_wall_center[t, x], to_units=tube_units["temperature"]
            )

        # Shell side wall temperature
        @self.Constraint(
            self.flowsheet().config.time,
            shell.length_domain,
            doc="shell side wall temperature",
        )
        def temp_wall_shell_eqn(b, t, x):
            return (
                b.total_heat_transfer_coeff_shell[t, x]
                * (shell.properties[t, x].temperature - b.temp_wall_shell[t, x])
                * (b.thickness_tube / 2 / b.therm_cond_wall + b.rfouling_shell)
                == b.temp_wall_shell[t, x] - b.temp_wall_center[t, x]
            )

        def heat_accumulation_term(b, t, x):
            return b.heat_accumulation[t, x] if b.config.dynamic else 0

        # Center point wall temperature based on energy balance for tube wall heat holdup
        @self.Constraint(
            self.flowsheet().config.time,
            tube.length_domain,
            doc="center point wall temperature",
        )
        def temp_wall_center_eqn(b, t, x):
            # heat_shell and heat_tube are positive when heat flows into those
            # control volumes (and out of the wall), hence the negative sign
            # on heat_accumulation_term
            return -heat_accumulation_term(b, t, x) == (
                b.heat_shell[t, x]
                * b.length_flow_shell
                / pyunits.convert(b.length_flow_tube, to_units=shell_units["length"])
                + pyunits.convert(
                    b.heat_tube[t, x],
                    to_units=shell_units["power"] / shell_units["length"],
                )
            )

        if not self.config.dynamic:
            z0 = shell.length_domain.first()
            z1 = shell.length_domain.last()

            @self.Expression(self.flowsheet().config.time)
            def total_heat_duty(b, t):
                enth_in = b.hot_side.properties[t, z0].get_enthalpy_flow_terms(p_hot)
                enth_out = b.hot_side.properties[t, z1].get_enthalpy_flow_terms(p_hot)

                return pyunits.convert(
                    enth_in - enth_out,  # Hot side loses enthalpy
                    to_units=shell_units["power"],  # Hot side isn't always the shell
                )

            @self.Expression(self.flowsheet().config.time)
            def log_mean_delta_temperature(b, t):
                dT0 = b.hot_side.properties[t, z0].temperature - pyunits.convert(
                    b.cold_side.properties[t, z0].temperature,
                    to_units=shell_units["temperature"],
                )
                dT1 = b.hot_side.properties[t, z1].temperature - pyunits.convert(
                    b.cold_side.properties[t, z1].temperature,
                    to_units=shell_units["temperature"],
                )
                return (dT0 - dT1) / log(dT0 / dT1)

            @self.Expression(self.flowsheet().config.time)
            def overall_heat_transfer_coefficient(b, t):
                return b.total_heat_duty[t] / (
                    b.total_heat_transfer_area * b.log_mean_delta_temperature[t]
                )

    def calculate_scaling_factors(self):
        def gsf(obj):
            return iscale.get_scaling_factor(obj, default=1, warning=True)

        def ssf(obj, sf):
            iscale.set_scaling_factor(obj, sf, overwrite=False)

        def cst(con, sf):
            iscale.constraint_scaling_transform(con, sf, overwrite=False)

        sgsf = iscale.set_and_get_scaling_factor
        if self.config.shell_is_hot:
            shell = self.hot_side
            tube = self.cold_side
            shell_units = (
                self.config.hot_side.property_package.get_metadata().derived_units
            )
        else:
            shell = self.cold_side
            tube = self.hot_side
            shell_units = (
                self.config.cold_side.property_package.get_metadata().derived_units
            )

        shell_has_pressure_change = hasattr(self, "deltaP_shell")

        heat_exchanger_common.scale_common(
            self,
            shell,
            shell_has_pressure_change,
            make_reynolds=True,
            make_nusselt=True,
        )

        sf_di_tube = iscale.get_scaling_factor(
            self.do_tube, default=1 / value(self.di_tube)
        )

        sf_area_per_length_shell = value(
            self.length_flow_shell
            / (
                pyunits.convert(self.length_flow_tube, to_units=shell_units["length"])
                * const.pi
                * self.do_tube
                * self.number_rows_per_pass
                * self.number_columns_per_pass
            )
        )

        for t in self.flowsheet().time:
            for z in shell.length_domain:
                # FIXME get better scaling later
                ssf(self.v_tube[t, z], 1 / 20)
                sf_flow_mol_tube = gsf(tube.properties[t, z].flow_mol)

                cst(self.v_tube_eqn[t, z], sf_flow_mol_tube)

                # FIXME should get scaling of N_Re from defining eqn
                sf_N_Re_tube = sgsf(self.N_Re_tube[t, z], 1e-4)

                sf_visc_d_tube = gsf(tube.properties[t, z].visc_d_phase["Vap"])
                cst(self.N_Re_tube_eqn[t, z], sf_N_Re_tube * sf_visc_d_tube)

                sf_k_tube = gsf(tube.properties[t, z].therm_cond_phase["Vap"])

                sf_N_Nu_tube = sgsf(self.N_Nu_tube[t, z], 1 / 0.023 * sf_N_Re_tube**0.8)
                cst(self.N_Nu_tube_eqn[t, z], sf_N_Nu_tube)

                sf_heat_transfer_coeff_tube = sgsf(
                    self.heat_transfer_coeff_tube[t, z],
                    sf_N_Nu_tube * sf_k_tube / sf_di_tube,
                )
                cst(
                    self.heat_transfer_coeff_tube_eqn[t, z],
                    sf_heat_transfer_coeff_tube * sf_di_tube,
                )

                sf_T_tube = gsf(tube.properties[t, z].temperature)
                ssf(self.temp_wall_tube[t, z], sf_T_tube)
                cst(self.temp_wall_tube_eqn[t, z], sf_T_tube)

                sf_Q_tube = gsf(tube.heat[t, z])
                cst(self.heat_tube_eqn[t, z], sf_Q_tube)

                sf_T_shell = gsf(shell.properties[t, z].temperature)
                ssf(self.temp_wall_shell[t, z], sf_T_shell)
                cst(self.temp_wall_shell_eqn[t, z], sf_T_shell)

                sf_conv_heat_transfer_coeff_shell = gsf(
                    self.conv_heat_transfer_coeff_shell[t, z]
                )
                s_Q_shell = sgsf(
                    shell.heat[t, z],
                    sf_conv_heat_transfer_coeff_shell
                    * sf_area_per_length_shell
                    * sf_T_shell,
                )
                cst(
                    self.heat_shell_eqn[t, z], s_Q_shell * value(self.length_flow_shell)
                )
                # Geometric mean is overkill for most reasonable cases, but it mitigates
                # damage done when one stream has an unset scaling factor
                ssf(self.temp_wall_center[t, z], (sf_T_shell * sf_T_tube) ** 0.5)
                cst(self.temp_wall_center_eqn[t, z], (sf_Q_tube * s_Q_shell) ** 0.5)

    def _get_performance_contents(self, time_point=0):
        var_dict = {}

        expr_dict = {}
        expr_dict["HX Area"] = self.total_heat_transfer_area
        expr_dict["Delta T Driving"] = self.log_mean_delta_temperature[time_point]
        expr_dict["Total Heat Duty"] = self.total_heat_duty[time_point]
        expr_dict["Average HX Coefficient"] = self.overall_heat_transfer_coefficient[
            time_point
        ]

        return {"vars": var_dict, "exprs": expr_dict}

    def _get_stream_table_contents(self, time_point=0):
        return create_stream_table_dataframe(
            {
                "Hot Inlet": self.shell_inlet,
                "Hot Outlet": self.shell_outlet,
                "Cold Inlet": self.tube_inlet,
                "Cold Outlet": self.tube_outlet,
            },
            time_point=time_point,
        )



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Tests for rate forms
"""

import pytest

from pyomo.environ import Block, ConcreteModel, Var, units as pyunits

from idaes.models.properties.modular_properties.base.generic_reaction import (
    GenericReactionParameterBlock,
    ConcentrationForm,
)
from idaes.models.properties.modular_properties.reactions.rate_forms import *

from idaes.core.util.testing import PhysicalParameterTestBlock
from idaes.core.util.misc import add_object_reference


@pytest.mark.unit
def test_power_law_rate_no_order():
    m = ConcreteModel()

    # # Add a test thermo package for validation
    m.pparams = PhysicalParameterTestBlock()
    m.thermo = m.pparams.build_state_block([1])

    m.rparams = GenericReactionParameterBlock(
        property_package=m.pparams,
        base_units={
            "time": pyunits.s,
            "mass": pyunits.kg,
            "amount": pyunits.mol,
            "length": pyunits.m,
            "temperature": pyunits.K,
        },
        rate_reactions={
            "r1": {
                "stoichiometry": {("p1", "c1"): -1, ("p1", "c2"): 2},
                "rate_form": power_law_rate,
                "concentration_form": ConcentrationForm.moleFraction,
            }
        },
    )

    # Create a dummy state block
    m.rxn = Block([1])
    add_object_reference(
        m.rxn[1], "phase_component_set", m.pparams._phase_component_set
    )
    add_object_reference(m.rxn[1], "params", m.rparams)
    add_object_reference(m.rxn[1], "state_ref", m.thermo[1])

    m.rxn[1].k_rxn = Var(["r1"], initialize=1)

    power_law_rate.build_parameters(
        m.rparams.reaction_r1, m.rparams.config.rate_reactions["r1"]
    )

    # Check parameter construction
    assert isinstance(m.rparams.reaction_r1.reaction_order, Var)
    assert len(m.rparams.reaction_r1.reaction_order) == 4
    for i, v in m.rparams.reaction_r1.reaction_order.items():
        try:
            stoic = m.rparams.config.rate_reactions.r1.stoichiometry[i]
        except KeyError:
            stoic = 0

        if stoic < 1:
            assert v.value == -stoic
        else:
            assert v.value == 0

    # Check reaction form
    rform = power_law_rate.return_expression(m.rxn[1], m.rparams.reaction_r1, "r1", 300)

    assert str(rform) == str(
        m.rxn[1].k_rxn["r1"]
        * m.thermo[1].mole_frac_phase_comp["p1", "c1"]
        ** m.rparams.reaction_r1.reaction_order["p1", "c1"]
    )


@pytest.mark.unit
def test_power_law_rate_with_order():
    m = ConcreteModel()

    # # Add a test thermo package for validation
    m.pparams = PhysicalParameterTestBlock()
    m.thermo = m.pparams.build_state_block([1])

    m.rparams = GenericReactionParameterBlock(
        property_package=m.pparams,
        base_units={
            "time": pyunits.s,
            "mass": pyunits.kg,
            "amount": pyunits.mol,
            "length": pyunits.m,
            "temperature": pyunits.K,
        },
        rate_reactions={
            "r1": {
                "stoichiometry": {("p1", "c1"): -1, ("p1", "c2"): 2},
                "rate_form": power_law_rate,
                "concentration_form": ConcentrationForm.moleFraction,
                "parameter_data": {
                    "reaction_order": {
                        ("p1", "c1"): 1,
                        ("p1", "c2"): 2,
                        ("p2", "c1"): 3,
                        ("p2", "c2"): 4,
                    }
                },
            }
        },
    )

    # Create a dummy state block
    m.rxn = Block([1])
    add_object_reference(
        m.rxn[1], "phase_component_set", m.pparams._phase_component_set
    )
    add_object_reference(m.rxn[1], "params", m.rparams)
    add_object_reference(m.rxn[1], "state_ref", m.thermo[1])

    m.rxn[1].k_rxn = Var(["r1"], initialize=1)

    power_law_rate.build_parameters(
        m.rparams.reaction_r1, m.rparams.config.rate_reactions["r1"]
    )

    # Check parameter construction
    assert isinstance(m.rparams.reaction_r1.reaction_order, Var)
    assert len(m.rparams.reaction_r1.reaction_order) == 4
    assert m.rparams.reaction_r1.reaction_order["p1", "c1"].value == 1
    assert m.rparams.reaction_r1.reaction_order["p1", "c2"].value == 2
    assert m.rparams.reaction_r1.reaction_order["p2", "c1"].value == 3
    assert m.rparams.reaction_r1.reaction_order["p2", "c2"].value == 4

    # Check reaction form
    rform = power_law_rate.return_expression(m.rxn[1], m.rparams.reaction_r1, "r1", 300)

    assert str(rform) == str(
        m.rxn[1].k_rxn["r1"]
        * (
            m.thermo[1].mole_frac_phase_comp["p1", "c1"]
            ** m.rparams.reaction_r1.reaction_order["p1", "c1"]
            * m.thermo[1].mole_frac_phase_comp["p1", "c2"]
            ** m.rparams.reaction_r1.reaction_order["p1", "c2"]
            * m.thermo[1].mole_frac_phase_comp["p2", "c1"]
            ** m.rparams.reaction_r1.reaction_order["p2", "c1"]
            * m.thermo[1].mole_frac_phase_comp["p2", "c2"]
            ** m.rparams.reaction_r1.reaction_order["p2", "c2"]
        )
    )



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
import numpy as np

from ..geometry import Parallelepiped, RectPrism
from ..atom import Atom


def readPointsAndAtomsFromPOSCAR(filename, ImpliedElems=None):
    Pts = []
    Atoms = []
    BBox = Parallelepiped.fromPOSCAR(filename)
    blnAtomToAdd = False  # Used to track
    Elem = None  # Used to store most recent element in the new format
    with open(filename, "r") as infile:
        CommentLine = infile.readline()
        GSLine = infile.readline().split()
        GS = float(GSLine[0])
        VxLine = infile.readline()
        VyLine = infile.readline()
        VzLine = infile.readline()
        ElementNamesOrCountsLine = infile.readline().split()
        if ElementNamesOrCountsLine[0].isdigit():
            if ImpliedElems is not None:
                Elems = ImpliedElems
                ElemCountsLine = ElementNamesOrCountsLine
            else:
                raise ValueError(
                    "Tried to read a POSCAR without explicit elements provided"
                )
        else:
            assert ElementNamesOrCountsLine[0].isalpha()
            ElemsLine = ElementNamesOrCountsLine
            Elems = [Atom(Elem) for Elem in ElemsLine]
            ElemCountsLine = infile.readline().split()
        ElemCounts = [int(ElemCount) for ElemCount in ElemCountsLine]
        CartesianOrDirectLine = infile.readline().split()
        CorDFlag = CartesianOrDirectLine[0][0].lower()
        blnIsCartesian = CorDFlag == "c" or CorDFlag == "k"
        for t, Elem in enumerate(Elems):
            for i in range(ElemCounts[t]):
                XYZLine = infile.readline().split()
                if blnIsCartesian:
                    P = GS * np.array(
                        [float(XYZLine[0]), float(XYZLine[1]), float(XYZLine[2])],
                        dtype=float,
                    )
                else:
                    # NOTE: The global scaling GS has already been taken in the BBox
                    #       so no need to multiply it here
                    Pfrac = np.array(
                        [float(XYZLine[0]), float(XYZLine[1]), float(XYZLine[2])],
                        dtype=float,
                    )
                    P = np.inner(np.array([BBox.Vx, BBox.Vy, BBox.Vz]).T, Pfrac)
                Pts.append(P)
                Atoms.append(Elem)
    return Pts, Atoms


def writeDesignToPOSCAR(
    D, filename, CommentLine=None, GS=None, BBox=None, Elems=None, blnUseDirect=True
):
    if GS is None:
        GS = 1.0
    if BBox is None:
        BBox = RectPrism.fromPointsBBox(D.Canvas.Points)
        BBox.scale(2.0)
    if Elems is None:
        Elems = D.NonVoidElems
    with open(filename, "w") as outfile:
        outfile.write("{}\n".format(CommentLine if CommentLine is not None else ""))
        outfile.write("{}\n".format(GS))
        outfile.write(
            "{} {} {}\n".format(BBox.Vx[0] / GS, BBox.Vx[1] / GS, BBox.Vx[2] / GS)
        )
        outfile.write(
            "{} {} {}\n".format(BBox.Vy[0] / GS, BBox.Vy[1] / GS, BBox.Vy[2] / GS)
        )
        outfile.write(
            "{} {} {}\n".format(BBox.Vz[0] / GS, BBox.Vz[1] / GS, BBox.Vz[2] / GS)
        )
        outfile.write("{}\n".format(" ".join([Elem.Symbol for Elem in Elems])))
        outfile.write(
            "{}\n".format(" ".join([str(D.Contents.count(Elem)) for Elem in Elems]))
        )
        outfile.write("{}\n".format("Direct" if blnUseDirect else "Cartesian"))
        for Elem in Elems:
            for i in range(len(D)):
                P = (
                    D.Canvas.Points[i]
                    if blnUseDirect
                    else BBox.getFractionalCoords(D.Canvas.Points[i])
                )
                if D.Contents[i] == Elem:
                    outfile.write("{} {} {}\n".format(P[0] / GS, P[1] / GS, P[2] / GS))
